"""
dtaidistance.subsequence
~~~~~~~~~~~~~~~~~~~~~~~~

(requires version 2.3.0 or higher)

Subsequence search

:author: Wannes Meert
:copyright: Copyright 2017-2021 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""

from .dtw import SubsequenceAlignment, subsequence_alignment
